#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/device.h>
#include <linux/interrupt.h>
#include <linux/dma-mapping.h>
#include <linux/delay.h>

#include <asm/hardware.h>
#include <asm/arch/pxa-regs.h>

#include <asm/errno.h>
#include <asm/io.h>
#include <asm/dma.h>
#include <asm/hardware.h>
#include "Barc_snd.h"

#include <sound/driver.h>
#include <sound/core.h>
#include <sound/pcm.h>

#include "dma.h"

#define DMA_SIZEMASK (0xFFFFF)


struct sDMABuffer primaryOutput;
struct sDMABuffer primaryInput;

#ifdef TDEV
extern struct tdev_link* outlink;
extern int silence;
extern struct tdev_link* llink;
extern int out_count;
void output_callback(pxa_dma_desc *chan, int chunk) {
	//printk("oc:%p %i\n",outlink,chunk);
	if (silence != 0) {
		while (1) {
			if (out_count >= PRE_BUFFER) { silence = 0; break; }
			int add_count = llink->active;
			if (add_count == 0) break;
			out_count += add_count;
			llink = llink->next;
		}
	}

	unsigned* buf = primaryOutput.virt + (chunk * DMA_SIZE);
	int count = 0;
	if (silence == 0) {
	while (count < DMA_SIZE) {
		int write = outlink->active;
		int written = 0;
		unsigned char* data = outlink->data + outlink->offset;
		//printk("  :%x %x %i\n",buf,data,write,outlink->offset);
		while (written < write) {
			//printk("  :%d %d\n",count,written);
			int bytes = write - written;
			if (bytes > ((DMA_SIZE - count) >> 1)) bytes = (DMA_SIZE - count) >> 1;
			//printk("  :%i\n",bytes);
			if (bytes > 0) {
				//printk("  :%x %i %x %i\n",(buf + (count >> 2)), count, (data + (written >> 2)), bytes);
				size_t samples = (size_t)(bytes >> 1);
				int16_t* s =  (int16_t*)(data + (written >> 2));
				int16_t* d =  (int16_t*)(buf + (count >> 2));
				while (samples--) {
					*(d++) = *(s);
					*(d++) = *(s++);
				}
				written += bytes;
				count += (bytes << 1);
			} else break;
		}
		if (written == 0) break;
		outlink->offset = written;
		written = write - written;
		outlink->active = written;
		//printk("  :w %i\n",written);
		if (written == 0) outlink = outlink->next;
	}
	}
	
	//printk("--:%d, %x %i\n",count,(buf + (count >> 2)),DMA_SIZE - count);
	if (count < DMA_SIZE) { silence = 1; out_count = 0; llink = outlink; memset((buf + (count >> 2)), 0, DMA_SIZE - count); }
}
#else
void output_callback(pxa_dma_desc *chan, int chunk) {
	if(primaryOutput.fill > primaryOutput.chunksize) {
		primaryOutput.fill -= primaryOutput.chunksize;
		primaryOutput.underrun = 0;
		//printk("CoolSound: OutputCB: chunk %d done. PrimaryOutput fill:%u\n", chunk, primaryOutput.fill);
	} else {
		//if(!primaryOutput.underrun) printk("CoolSound: Warning: Primary output buffer underrun.\n");
		primaryOutput.fill = 0;
		primaryOutput.underrun = 1;
	}

	if(primaryOutput.callback) primaryOutput.callback(DMAP_PLAY, chunk);
}
#endif

void input_callback(pxa_dma_desc *chan, int chunk) {
	if((primaryInput.size-primaryInput.fill) > primaryInput.chunksize) {
		primaryInput.fill += primaryInput.chunksize;
		primaryInput.overflow = 0;
		//printk("CoolSound: InputCB: chunk %d done. PrimaryInput fill:%u\n", chunk, primaryInput.fill);
	} else {
		if(!primaryInput.overflow) printk("CoolSound: Warning: Primary input buffer overflow.\n");
		primaryInput.fill = primaryInput.size;
		primaryInput.overflow = 1;
	}


	if(primaryInput.callback) primaryInput.callback(DMAP_RECORD, chunk);
}

void DMABuffer_free(struct sDMABuffer *pBuffer) {
	//return;
	if(!pBuffer) return;

	if(pBuffer->flags & DBF_ALLOCATED) {
		dma_free_coherent(NULL, pBuffer->dmadesc_size, pBuffer->dmadesc_virt, pBuffer->dmadesc_phys);
		dma_free_coherent(NULL, pBuffer->size, pBuffer->virt, pBuffer->phys);
	}

	pBuffer->size = 0;
	pBuffer->dmadesc_size = 0;
	pBuffer->dmadesc_virt = NULL;
	pBuffer->dmadesc_phys = 0;
	pBuffer->virt = NULL;
	pBuffer->phys = 0;

	pBuffer->flags &= ~DBF_ALLOCATED;
}

#define MAX_DMA_XFER_SIZE	4096

int DMABuffer_allocate(struct sDMABuffer *pBuffer, size_t chunksize, int chunks) {
	pxa_dma_desc *dma_dring;

	size_t d_size = sizeof(pxa_dma_desc);
	unsigned int dma_dpp; /* dma descriptors per period */
	size_t dma_xfer_size = (chunksize >= MAX_DMA_XFER_SIZE) ?
			       MAX_DMA_XFER_SIZE : chunksize;
	dma_dpp = (chunksize - 1) / dma_xfer_size + 1;


	if(!pBuffer) return -EINVAL;
	if(chunks < 3) return -EINVAL;

	//Calling this function is save in any case
	DMABuffer_free(pBuffer);

	pBuffer->size = chunks * chunksize;

	//Check if requested size is too big for DMA
	if(pBuffer->size > DMA_SIZEMASK) return -EINVAL;

	pBuffer->virt = dma_alloc_coherent(NULL, pBuffer->size, &pBuffer->phys, GFP_KERNEL|GFP_DMA);

	if (!pBuffer->phys) {
		printk("Failed to allocate PCM buffer\n");
		return -ENOMEM;
	}

	pBuffer->dmadesc_size = chunks * dma_dpp * d_size + 15;
	pBuffer->dmadesc_virt = dma_alloc_coherent(NULL, pBuffer->dmadesc_size, &pBuffer->dmadesc_phys, GFP_KERNEL|GFP_DMA);
	
	if (!pBuffer->dmadesc_virt) {
		printk("Failed to alloc DMA buffer of size %d\n", pBuffer->dmadesc_size);
		dma_free_coherent(NULL, pBuffer->size, pBuffer->virt, pBuffer->phys);
		
		pBuffer->size = 0;
		pBuffer->virt = NULL;
		pBuffer->phys = 0;
		pBuffer->dmadesc_size = 0;
		return -ENOMEM;
	}

	memset(pBuffer->virt, 0x42, pBuffer->size);

	/* setup the descriptors ring pointer, aligned on a 16 bytes boundary */
	dma_dring = ((dma_addr_t) pBuffer->dmadesc_virt & 0xf) ?
			 (pxa_dma_desc *) (((dma_addr_t) pBuffer->dmadesc_virt & ~0xf) + 16) :
			 (pxa_dma_desc *) pBuffer->dmadesc_virt;
	pBuffer->dma_dring_p = pBuffer->dmadesc_phys +
			  (dma_addr_t) dma_dring - (dma_addr_t) pBuffer->dmadesc_virt;

	{
		unsigned int i, j, di;
		dma_addr_t p, d_buf;
		
		/* fill the descriptors ring */
		di = 0; /* current descriptor index */
		p = pBuffer->phys;
		
		/* iterate over periods */
		for (i = 0; i < chunks; i++, p += chunksize) {
			d_buf = p; /* d_buf: address in buffer for current dma descriptor */
			/* iterate over dma descriptors in a period */
			for (j = 0; j < dma_dpp; j++, di++, d_buf += dma_xfer_size) {
				/* link to next descriptor */
				dma_dring[di].ddadr = pBuffer->dma_dring_p + (di + 1) * d_size;
				dma_dring[di].dsadr = d_buf;
				dma_dring[di].dtadr = __PREG(SADR);
				dma_dring[di].dcmd = DCMD_INCSRCADDR | DCMD_FLOWTRG |
					DCMD_BURST32 | DCMD_WIDTH4;
				dma_dring[di].dcmd |=
					((p + chunksize - d_buf) >= dma_xfer_size) ?
					dma_xfer_size : p + chunksize - d_buf; /* transfer length */
			}
			dma_dring[di - 1].dcmd |= DCMD_ENDIRQEN; /* period irq */
		}
		dma_dring[di - 1].ddadr = pBuffer->dma_dring_p; /* close the ring */	
	}


	pBuffer->chunksize = chunksize;
	pBuffer->chunks = chunks;
	pBuffer->fill	= 0;
	pBuffer->flags |= DBF_ALLOCATED;

	return 0;
}

void DMABuffer_disassociate(struct sDMABuffer *pBuffer) {
	if(!pBuffer) return;


	if(pBuffer->flags & DBF_ASSOCIATED)
		pxa_free_dma(pBuffer->channel);

	pBuffer->flags &= ~(DBF_ASSOCIATED | DBF_RUNNING);
}

static void dma_irq(int ch, void *data, struct pt_regs *regs)
{
	u32 dcsr;
	struct sDMABuffer *pBuffer = (struct sDMABuffer *)data;

	dcsr = DCSR(ch);
	DCSR(ch) = dcsr & ~DCSR_STOPIRQEN;

	if (dcsr & DCSR_BUSERR) {
		return;
	}

	if (dcsr & DCSR_ENDINTR) {
		pBuffer->cchunk++;
		if(pBuffer->cchunk > pBuffer->chunks)
			pBuffer->cchunk = 0;

		output_callback(&pBuffer->client, pBuffer->cchunk);
	}
}

int DMABuffer_associate(struct sDMABuffer *pBuffer, enum eDMAPurpose purpose) {
	int ret;
	if(!pBuffer) return -EINVAL;

	DMABuffer_disassociate(pBuffer);

	if(purpose == DMAP_PLAY) {
		if((ret = pxa_request_dma("i2s out", DMA_PRIO_LOW, dma_irq, pBuffer)) < 0) {
			printk("Failed to request dma\n");
			return -EBUSY;
		}
		pBuffer->channel = ret;
		DRCMRTXSADR = pBuffer->channel | DRCMR_MAPVLD;
		pBuffer->flags |= DBF_ASSOCIATED;
		return 0;
	}

	if(purpose == DMAP_RECORD) {
		if((ret = pxa_request_dma("i2s in", DMA_PRIO_LOW, dma_irq, pBuffer)) < 0) {
			printk("Failed to request dma\n");
			return -EBUSY;
		}
		pBuffer->channel = ret;
		DRCMRRXSADR = pBuffer->channel | DRCMR_MAPVLD;
		pBuffer->flags |= DBF_ASSOCIATED;
		return 0;
	}

	return -EINVAL;
}


int DMABuffer_run(struct sDMABuffer *pBuffer) {
	pr_debug("DMABufferRun [%p]\n",pBuffer);
	if(!pBuffer) return -EINVAL;

	if(!(pBuffer->flags & DBF_ALLOCATED)) return -EBUSY;
	if(!(pBuffer->flags & DBF_ASSOCIATED)) return -EBUSY;

	if(pBuffer->flags & DBF_RUNNING) return 0;
	pr_debug("DMABufferRun2 [%p]\n",pBuffer);

	pBuffer->flags |= DBF_RUNNING;
	
	pBuffer->cchunk = 0;
	DDADR(pBuffer->channel) = pBuffer->dma_dring_p;
	DCSR(pBuffer->channel) = DCSR_RUN;

	pr_debug("Buffer[%p] now running.\n", pBuffer);
	return 0;
}

int DMABuffer_running(struct sDMABuffer *pBuffer) {
	if(!pBuffer) return 0;
	return (pBuffer->flags & DBF_RUNNING);
}

int DMABuffer_ready(struct sDMABuffer *pBuffer) {
	if(!pBuffer) return 0;
	return (pBuffer->flags & ( DBF_ASSOCIATED | DBF_ALLOCATED ));
}

int DMABuffer_getRWPos(struct sDMABuffer *pBuffer) {
	if(!pBuffer) return -EINVAL;

	if(!(pBuffer->flags & DBF_ALLOCATED)) return -EBUSY;
	if(!(pBuffer->flags & DBF_ASSOCIATED)) return -EBUSY;

	if(!(pBuffer->flags & DBF_RUNNING)) {
		return 0;
	}

	return pBuffer->cchunk * pBuffer->chunksize;
}

void DMABuffer_stop(struct sDMABuffer *pBuffer) {
	unsigned int timeout = 100; /* safety timeout */

	if(!pBuffer) return;

	if(!(pBuffer->flags & DBF_ASSOCIATED)) return;

	DCSR(pBuffer->channel) = DCSR_STOPIRQEN;

	while (!(DCSR(pBuffer->channel) & DCSR_STOPSTATE) && timeout > 0) {
		 udelay(10);
		 timeout--;
	}

	pBuffer->flags &= (~DBF_RUNNING);
	//printk("Buffer[%p] now stopped.\n");
}


int sound_dma_init(sound_dma_callback_t callback) {
	int res;

	//Setup device pointer
	primaryOutput.flags = 0;
	primaryInput.flags = 0;

	primaryOutput.callback = callback;
	primaryInput.callback = callback;

	//Associate buffers
	if((res = DMABuffer_associate(&primaryOutput, DMAP_PLAY))) return res;
	if((res = DMABuffer_associate(&primaryInput, DMAP_RECORD))) return res;

	printk("CoolSound: sound_dma_init: Buffers allocated and associated.\n");
	return 0;
}

int sound_dma_suspend() {
	//TODO: remeber which channels were running,
	//      shut them down safely and restart them on resume

	return 0;
}

int sound_dma_resume() {
	int res = 0;

	//DMA suspend/resume is broken - we need to reassociate buffers
	if((res = DMABuffer_associate(&primaryOutput, DMAP_PLAY))) return res;
	if((res = DMABuffer_associate(&primaryInput, DMAP_RECORD))) return res;

	return res;
}

int sound_dma_setup_buffers(int chunksize, int chunks) {
	int res;

	//Allocate buffers
	if((res = DMABuffer_allocate(&primaryOutput, chunksize, chunks))) return res;
	if((res = DMABuffer_allocate(&primaryInput, chunksize, chunks))) return res;

	return 0;
}

int sound_dma_output_running(void) {
	return DMABuffer_running(&primaryOutput);
}

int sound_dma_input_running(void) {
	return DMABuffer_running(&primaryInput);
}

int sound_dma_output_ready(void) {
	return DMABuffer_ready(&primaryOutput);
}

int sound_dma_input_ready(void) {
	return DMABuffer_ready(&primaryInput);
}

int sound_dma_output_start(void) {
	return DMABuffer_run(&primaryOutput);
}

void sound_dma_output_stop(void) {
	DMABuffer_stop(&primaryOutput);
}

int sound_dma_output_pos(void) {
	return DMABuffer_getRWPos(&primaryOutput);
}

int sound_dma_input_start(void) {
	printk("Tried to enable sound input but disabled so far\n");
	return 0;
	//return DMABuffer_run(&primaryInput);
}

void sound_dma_input_stop(void) {
	printk("Tried to stop sound input but disabled so far\n");
//	DMABuffer_stop(&primaryInput);
}

int sound_dma_input_pos(void) {
	return DMABuffer_getRWPos(&primaryInput);
}

int sound_dma_output_allocated(void) {
	return (primaryOutput.flags & DBF_ALLOCATED);
}

size_t sound_dma_output_size(void) {
	return (primaryOutput.size);
}

void *sound_dma_output_phys(void) {
	return (void *)(primaryOutput.phys);
}

int sound_dma_input_allocated(void) {
	return (primaryInput.flags & DBF_ALLOCATED);
}

size_t sound_dma_input_size(void) {
	return (primaryInput.size);
}

void *sound_dma_input_phys(void) {
	return (void *)(primaryInput.phys);
}

void sound_dma_output_reset(void) {
	primaryOutput.pointer 	= 0;
	primaryOutput.fill	= 0;
	primaryOutput.underrun	= 0;
	primaryOutput.overflow	= 0;

}

void sound_dma_input_reset(void) {
	primaryInput.pointer 	= 0;
	primaryInput.fill	= 0;
	primaryInput.underrun	= 0;
	primaryInput.overflow	= 0;
}

int sound_dma_output_free(void) {
	int res;
/*
	int rp, wp;

	rp = DMABuffer_getRWPos(&primaryOutput);
	wp = primaryOutput.pointer;

	if(rp>wp) 	res =  (rp - wp) - 1;
	else 		res = primaryOutput.size - (wp-rp) - 1;
*/

	res = primaryOutput.size - primaryOutput.fill;

	res>>=2;
	//printk("sound_dma_output_free() wp:%d rp:%d free:%d\n",wp, rp, res);
	return res;
}

int sound_dma_input_filled(void) {
	int res;
/*
	int rp, wp;

	wp = DMABuffer_getRWPos(&primaryInput);
	rp = primaryInput.pointer;

	if(wp>=rp) 	res =  (wp - rp);
	else 		res = primaryInput.size - (rp-wp);
*/

	res = primaryInput.fill;

	res>>=2;
	//printk("sound_dma_input_filled() wp:%d rp:%d fill:%d\n",wp, rp, res);
	return res;
}


void sound_dma_output_zeroes(int samples, int offset) {
	int pointer;

	if(samples > (primaryOutput.size>>2)) {
		printk("CoolSound: Warning: sound_dma_output_zero() samples(%i) > buffersize(%u).\n",
			samples, (primaryOutput.size>>2));

		samples = primaryOutput.size >> 2;
	}

	pointer = (primaryOutput.pointer+(offset<<2)) % primaryOutput.size;

	if((primaryOutput.size - pointer) >= (samples<<2)) {
		memset(primaryOutput.virt + pointer, 0, samples<<2);
	} else {
		memset(primaryOutput.virt + pointer, 0, (primaryOutput.size - pointer));
		samples-=(primaryOutput.size - pointer)>>2;
		memset(primaryOutput.virt, 0, samples<<2);
	}
}

void sound_dma_output_put(void *pData, int size, int offset, cs_primixer_t primixer, unsigned int flags) {
	int pointer;

	if (size <= 0)
		return;

	if (pData == NULL) {
		printk(KERN_WARNING "%s: NULL data pointer\n", __FUNCTION__);
		return;
	}

	if ((size<<2) > primaryOutput.size) {
		printk(KERN_WARNING "%s: Invalid size %d\n", __FUNCTION__, size);
		return;
	}

	pointer = (primaryOutput.pointer+(offset<<2)) % primaryOutput.size;


	if((primaryOutput.size - pointer) >= (size<<2)) {
		primixer(primaryOutput.virt + pointer, pData, size, flags);
	} else {
		primixer(primaryOutput.virt + pointer, pData, (primaryOutput.size - pointer)>>2, flags);
		size-=(primaryOutput.size - pointer)>>2;
		pData+=(primaryOutput.size - pointer)>>2;
		primixer(primaryOutput.virt, pData, size, flags);
	}

}

void sound_dma_input_get(void *pData, int size, int offset, cs_primixer_t primixer, unsigned int flags) {
	int pointer;

	if (size <= 0)
		return;

	if (pData == NULL) {
		printk(KERN_WARNING "%s: NULL data pointer\n", __FUNCTION__);
		return;
	}

	if ((size<<2) > primaryInput.size) {
		printk(KERN_WARNING "%s: Invalid size %d\n", __FUNCTION__, size);
		return;
	}

	pointer = (primaryInput.pointer+(offset<<2)) % primaryInput.size;

	if((primaryInput.size - pointer) >= (size<<2)) {
		primixer(pData, primaryInput.virt + pointer, size, flags);
	} else {
		primixer(pData, primaryInput.virt + pointer, (primaryInput.size - pointer)>>2, flags);
		size 	-= (primaryInput.size - pointer)>>2;
		pData	+= (primaryInput.size - pointer)>>2;
		primixer(pData, primaryInput.virt, size, flags);
	}
}

void sound_dma_input_advance(int size) {
	unsigned long flags;
	
	primaryInput.pointer = (primaryInput.pointer + (size<<2)) % primaryInput.size;
	
	local_irq_save(flags);
	if(primaryInput.fill < (size<<2)) {
		printk("CoolSound: Warning: primary Input buffer underrun!\n");
		primaryInput.fill = 0;
	} else {
		primaryInput.fill -= (size<<2);
	}
	local_irq_restore(flags);
}

void sound_dma_output_advance(int size) {
	unsigned long flags;

	primaryOutput.pointer = (primaryOutput.pointer + (size<<2)) % primaryOutput.size;

	local_irq_save(flags);
	primaryOutput.fill += (size<<2);
	if(primaryOutput.fill > primaryOutput.size) {
		printk("CoolSound: Warning: primary Output buffer overflow!\n");
		primaryOutput.fill %= primaryOutput.size;
	}
	local_irq_restore(flags);
}

void sound_dma_filter_init() {}
void sound_dma_filter_shutdown() {}
void sound_dma_filter_resync() {}
void sound_dma_filter_magic() {}
